#include <iostream>
#include <sstream>
#include <vector>
#include <Eigen/Core>
#include <Eigen/Geometry>
#include <Eigen/StdVector>

#include "fileIO.hpp"

using namespace std;
using namespace Eigen;

#ifndef _MATRIX_
#define _MATRIX_
typedef Matrix<double, 3, 4> Matrix34d;
#endif

// constants
const double EPS = 1e-10;
const Matrix3d I3 = Matrix3d::Identity();

namespace bundleadjustment
{
  class BundleAdjustment
  {
  private:
    int pointN;     // # of 3-D points
    int cameraN;    // # of cameras
    double f0;      // scaler
    double c;       // constant for Levenberg-Marquardt

    int In;            // # of visibule points
    int** I = NULL;    // 1 if the point `a' is vibible in the image `ｋ'
    double** x = NULL; // x
    double** y = NULL; // y

    // variables
    Vector3d* X = NULL;;        // 3-D points
    Matrix34d* cameraMat = NULL;// camera matrices
    double*f = NULL;            // focal lengths
    double*u = NULL;            // principal points: u
    double*v = NULL;            // principal points: v
    Vector3d* t = NULL;         // translations
    Matrix3d* R = NULL;         // rotations


    // error variables
    double error;       // reprojection error
    double errorTilde;  // reprojection error
    double* derror;     // error of 1st derivatives

    // for 1st and 2nd derivatives
    double** p = NULL;
    double** q = NULL;
    double** r = NULL;

    // 2nd derivatives
    Matrix3d* hE = NULL;
    MatrixXd* hF = NULL;
    MatrixXd hG;

    // variables for solution
    VectorXd deltaXiP;
    VectorXd deltaXiF;

    // 1st derivatives
    VectorXd dP;
    VectorXd dF;

  public:
    EIGEN_MAKE_ALIGNED_OPERATOR_NEW  // need this macro for Eigen members
    BundleAdjustment();
    ~BundleAdjustment();


    void calcpqr();
    void calcError();

    Vector3d calcdpqrak(int alpha, int kappa, int l);
    void calcdError();

    void calcddError();
    void calchE();
    void calchF();
    void calchG();
    double calchkl(int k, int l, int alpha, int kappa);
    double calchEkl(int k, int l, int alpha);
    double calchFkl(int k, int l);
    double calchGkl(int k, int l);
    int offset(int k);

    /*solve*/
    void solveEquations();
    void setdPdF();
    void solveEquationsdXiF();
    void solveEquationsdXiP();

    /*tilde*/
    void calcErrorTilde();
    Vector3d calcpqrTildeak(int alpha, int kappa);
    double getdeltaXiF(int kappa, int param);
    Matrix3d getRotateMat(const Vector3d& l);

    void renewParams();

    /**/
    void bundleAdjustment(
			  string file_xy,
			  string file_3d,
			  string file_K,
			  string file_R,
			  string file_t,
			  int width,
			  int height,
			  double f0);
    void setParams(
		   string file_xy,
		   string file_3d,
		   string file_K,
		   string file_R,
		   string file_t,
		   int width,
		   int height,
		   double f0);
    void init();


    //set
    void setX(int alpha, const Vector3d& X);
    void setf(int kappa, double f);
    void setu(int kappa, double u);
    void setv(int kappa, double v);
    void sett(int kappa, const Vector3d& t);
    void setR(int kappa, const Matrix3d& R);
    void setCameraMat(int kappa, const Matrix34d& P);

  };
}
