#include <iostream>
#include <vector>
#include <Eigen/Core>
#include <Eigen/Geometry>
#include <Eigen/StdVector>

#include "BundleAdjustment.hpp"
#include "fileIO.hpp"

using namespace std;
using namespace Eigen;
using namespace bundleadjustment;

#ifndef _MATRIX_
#define _MATRIX_
typedef Matrix<double, 3, 4> Matrix34d;
#endif

//
// set parameters
//
void BundleAdjustment::setParams(
   string file_xy,
   string file_3d,
   string file_K,
   string file_R,
   string file_t,
   int width,
   int height,
   double f0
)
{
   // Read files
   vector< vector<Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > > data_xy;
   vector<Vector3d,Eigen::aligned_allocator<Eigen::Vector3d> > data_3d;
   vector<Matrix3d,Eigen::aligned_allocator<Eigen::Matrix3d> > data_R;
   vector<Matrix3d,Eigen::aligned_allocator<Eigen::Matrix3d> > data_K;
   vector<Vector3d,Eigen::aligned_allocator<Eigen::Vector3d> > data_t;

   readDataxy(data_xy, file_xy);
   readDataVec3d(data_3d, file_3d);
   readDataMat3d(data_K, file_K);
   readDataMat3d(data_R, file_R);
   readDataVec3d(data_t, file_t);

   int N = data_xy.size();
   int M = data_xy[0].size();
   this->f0 = f0;
   this->pointN = N;
   this->cameraN = M;
   cout << "# of cameras = " << M << ",  # of points = " << N << endl;

   // Allocate
   init();

   // cout << "set xy" << endl;

   // insert x and y, and count In
   // origin: upper left

   int id;
   int cnti = 0;
   for(int kappa = 0; kappa < M; kappa++){
      id = kappa;
      for(int alpha = 0; alpha < N; alpha++){
         this->x[kappa][alpha] = -data_xy[alpha][id](1) + height/2;
         this->y[kappa][alpha] =  data_xy[alpha][id](0) - width/2;
         if(data_xy[alpha][id](0) >= 0){
            this->I[kappa][alpha] = 1;
            cnti++;
         }else{
            this->I[kappa][alpha] = 0;
         }
      }
   }
   this->In = cnti;


   // normalization
   // cout << "normalization" << endl;
   Vector3d t1 = data_t[0];
   Vector3d t2 = data_t[1];
   Matrix3d R1 = data_R[0];
   Vector3d j;

   j << 0, 1, 0;
   double s = j.dot(R1.transpose() * (t2 -t1));

   // t
   for(int i = 0; i < M; i++){
      data_t[i] = 1.0 / s * R1.transpose() * (data_t[i]- t1);
      sett(i, data_t[i]);
   }

   // Matrix3d R
   for(int i = 0; i < M; i++){
      data_R[i] = R1.transpose() * data_R[i];
      setR(i, data_R[i]);
   }

   for(int i = 0; i < N; i++){
      data_3d[i] = 1.0 / s * R1.transpose() * (data_3d[i] - t1);
      setX(i, data_3d[i]);
   }

   double u, v, f;
   Matrix34d P;

   for(int i = 0; i < M; i++){
      f = data_K[i](0, 0);
      u = data_K[i](0, 2);
      v = data_K[i](1, 2);
      setf(i, f);
      setu(i, u);
      setv(i, v);
      P << data_K[i] * data_R[i].transpose(), -data_K[i] * data_R[i].transpose() * data_t[i];
      setCameraMat(i, P);
   }

   cout << "output ply of initial data: init_3d.ply" << endl;
   outDataPly("init_3d.ply", data_3d.data(), data_3d.size());
}

//
//  Procecure 11.1 bundle adjustment
//
void BundleAdjustment::bundleAdjustment(
   string file_xy,
   string file_3d,
   string file_K,
   string file_R,
   string file_t,
   int width,
   int height,
   double f0)
{
   double e = 0.0;       // reprojection error
   double eTilde = 0.0;  // reprojection error
   //double delta = this->In * pow(0.01, 2) / pow(this->f0, 2);  // threshold for exiting loop
   double delta = 0.0001;
   cout << "delta = " << delta << endl;

   // step 1
   setParams(file_xy, file_3d, file_K, file_R, file_t, width, height, f0);    // initialization
   calcpqr();      // compute pqr
   calcError();    // compute reprojection error
   e = this->f0 * sqrt(this->error / (2*this->In - (3*this->pointN + 9*this->cameraN - 7)) );
   this->c = 0.0001;

   cout << "e = " << e << ", ";
   cout << "c = " << this->c << "\n";

   int loop = 0;   // loop count
   while (true) {
      cout << "loop: " << loop << "\n";

      // step 2．
      calcdError();   // compute 1st derivatives
      calcddError();  // compute 2nd derivatives


      do {
         // step 3
         solveEquations();   // solve simultanous equation

         // step 4 and 5
         calcErrorTilde();   // compute ``tilde'' variables

         eTilde = this->f0 * sqrt(this->errorTilde / (2*this->In - (3*this->pointN + 9*this->cameraN - 7)));
         cout << "e = " << e << ", ";
         cout << "~e = " << eTilde << "\n";

         if ( eTilde > e ) {
            this->c = 10.0 * this->c;
            cout << "c = " << this->c << "\n";
            calcdError();
            calcddError();  // compute 2nd derivatives
         }
         // if tilde(E) > E then c <= 10c, and go to step 3
      } while ( eTilde > e );

      // step 6.
      renewParams();  // update parameters
      calcpqr();
      calcError();

      // output debug information
      {
         eTilde = this->f0 * sqrt(this->errorTilde / (2*this->In - (3*this->pointN + 9*this->cameraN - 7)));
         cout << "e = " << e << ", ";
         cout << "~e = " << eTilde << "\n";


         loop++;
         vector<Matrix3d,Eigen::aligned_allocator<Eigen::Matrix3d> > R;
         vector<Vector3d,Eigen::aligned_allocator<Eigen::Vector3d> > K;
         vector<Vector3d,Eigen::aligned_allocator<Eigen::Vector3d> > t;
         for (int i = 0; i < this->cameraN; i++) {
            R.push_back(this->R[i]);
            K.push_back(Vector3d(this->f[i], this->u[i], this->v[i]));
            t.push_back(this->t[i]);
         }

         // output temporary shape
         if (!(loop % 20))
         {
            ostringstream stream;
            stream << loop;
            string result = stream.str();
            string filename = "bundle_out/loop_" + result + ".ply";
            cout << "output: " << filename << endl;
            outDataPly(filename, this->X, this->pointN);
            string out_file;
            out_file = "bundle_out/loop_" + result + "_R";
            out_Mat3d(R, out_file);
            out_file = "bundle_out/loop_" + result + "_K";
            out_Vec3d(K, out_file);
            out_file = "bundle_out/loop_" + result + "_t";
            out_Vec3d(t, out_file);
         }
      }

      // if |tilde(E) - E| > delta, then E <= tilde(E), c <= c/10, and go to step 2
      if ( abs(eTilde - e) <= delta ) {
         break;
      } else {
         e = eTilde;
         this->c = this->c / 10.0;
         cout << "c = " << this->c << "\n";
      }
   }

   // output the final result
   cout << "output ply of final data: final_3d.ply" << endl;
   outDataPly("final_3d.ply", this->X, this->pointN);
}


// Eq.(11.12)
void BundleAdjustment::calcpqr()
{
   Matrix34d P;
   Vector3d pqr;
   Vector4d XYZ1;
   int alpha, kappa;
   int N = this->pointN;
   int M = this->cameraN;

   for(kappa = 0; kappa < M; kappa++){
      P << this->cameraMat[kappa];
      for(alpha = 0; alpha < N; alpha++){
         XYZ1 << this->X[alpha], 1.0;
         pqr = P * XYZ1;
         this->p[kappa][alpha] = pqr(0);
         this->q[kappa][alpha] = pqr(1);
         this->r[kappa][alpha] = pqr(2);
      }
   }
}

// Eq.(11.13)
void BundleAdjustment::calcError()
{

   int N = this->pointN;
   int M = this->cameraN;
   double f0 = this->f0;
   double e = 0;
   double sqrt1, sqrt2;
   int alpha, kappa;

   for(kappa = 0; kappa < M; kappa++){
      for(alpha = 0; alpha < N; alpha++){
         if(this->I[kappa][alpha] == 1){
            sqrt1 = this->p[kappa][alpha] / this->r[kappa][alpha] - this->x[kappa][alpha] / f0;
            sqrt2 = this->q[kappa][alpha] / this->r[kappa][alpha] - this->y[kappa][alpha] / f0;
            e += pow(sqrt1, 2) + pow(sqrt2, 2);
         }
      }
   }
   this->error = e;
}

// Eqs.(11.16), (11.17), (11.18), and (11.20)
Vector3d BundleAdjustment::calcdpqrak(int alpha, int kappa, int l)
{
   int N3 = this->pointN * 3;
   int M9 = this->cameraN * 9;
   Vector3d dpqr = Vector3d::Zero();
   int beta, lambda;
   Matrix34d P;
   int id;
   Vector3d r_k1, r_k2, r_k3;
   Matrix3d R;

   if(l < N3){
      // 3-D points
      beta = l / 3;
      if(beta != alpha){
         return dpqr;
      }
      P = this->cameraMat[kappa];
      id = l % 3;
      dpqr(0) = P(0, id);
      dpqr(1) = P(1, id);
      dpqr(2) = P(2, id);

   }else{
      // camera parameters
      lambda = (l - N3) / 9;
      if(lambda != kappa){
         return dpqr;
      }

      id = (l - N3) % 9;
      switch (id) {
        case 0:
           dpqr(0) = (this->p[kappa][alpha] - this->u[kappa] / this->f0 * this->r[kappa][alpha]) / this->f[kappa];
           dpqr(1) = (this->q[kappa][alpha] - this->v[kappa] / this->f0 * this->r[kappa][alpha]) / this->f[kappa];
           dpqr(2) = 0;
           break;
        case 1:
           dpqr(0) = this->r[kappa][alpha] / this->f0;
           dpqr(1) = 0;
           dpqr(2) = 0;
           break;
        case 2:
           dpqr(0) = 0;
           dpqr(1) = this->r[kappa][alpha] / this->f0;
           dpqr(2) = 0;
           break;
        case 3:
        case 4:
        case 5:
           id = id - 3;
           R = this->R[kappa];
           r_k1 << R(0, 0), R(1, 0), R(2, 0);
           r_k2 << R(0, 1), R(1, 1), R(2, 1);
           r_k3 << R(0, 2), R(1, 2), R(2, 2);
           dpqr(0) = -(this->f[kappa] * r_k1(id) + this->u[kappa] * r_k3(id));
           dpqr(1) = -(this->f[kappa] * r_k2(id) + this->v[kappa] * r_k3(id));
           dpqr(2) = -this->f0 * r_k3(id);
           break;
        case 6:
        case 7:
        case 8:
           id = id - 6;
           R = this->R[kappa];
           r_k1 << R(0, 0), R(1, 0), R(2, 0);
           r_k2 << R(0, 1), R(1, 1), R(2, 1);
           r_k3 << R(0, 2), R(1, 2), R(2, 2);
           Vector3d vec1, vec2;
           vec2 = this->X[alpha] - this->t[kappa];
           vec1 = this->f[kappa] * r_k1 + this->u[kappa] * r_k3;
           dpqr(0) = (vec1.cross(vec2))(id);
           vec1 = this->f[kappa] * r_k2 + this->v[kappa] * r_k3;
           dpqr(1) = (vec1.cross(vec2))(id);
           vec1 = this->f0 * r_k3;
           dpqr(2) = (vec1.cross(vec2))(id);
           break;
      }
   }
   return dpqr;
}

// Eq.(11.14)
void BundleAdjustment::calcdError()
{
   int N = this->pointN;
   int M = this->cameraN;
   int N3 = this->pointN * 3;
   int M9 = this->cameraN * 9;
   int alpha, kappa;
   Vector3d dpqr;
   Vector3d pqr;
   double eql1, eql2, eqr1, eqr2;
   double e;

   // Points
   for(int l = 0; l < N3; l++){
      e = 0;
      alpha = l / 3;
      for(kappa = 0; kappa < M; kappa++){
         if(this->I[kappa][alpha]){
            dpqr = calcdpqrak(alpha, kappa, l);
            pqr << this->p[kappa][alpha], this->q[kappa][alpha], this->r[kappa][alpha];
            eql1 = pqr(0) / pqr(2) - this->x[kappa][alpha] / this->f0;
            eqr1 = pqr(1) / pqr(2) - this->y[kappa][alpha] / this->f0;
            eql2 = pqr(2) * dpqr(0) - pqr(0) * dpqr(2);
            eqr2 = pqr(2) * dpqr(1) - pqr(1) * dpqr(2);
            e += (eql1 * eql2 + eqr1 * eqr2) / pow(pqr(2), 2);
         }
      }
      this->derror[l] = 2 * e;
   }

   // camera parameters
   int N3_M9 = N3 + M9;
   for(int l = N3; l < N3_M9; l++){
      e = 0;
      kappa = (l - N3) / 9;
      for(alpha = 0; alpha < N; alpha++){
         if(this->I[kappa][alpha]){
            dpqr = calcdpqrak(alpha, kappa, l);
            pqr << this->p[kappa][alpha], this->q[kappa][alpha], this->r[kappa][alpha];
            eql1 = pqr(0) / pqr(2) - this->x[kappa][alpha] / this->f0;
            eqr1 = pqr(1) / pqr(2) - this->y[kappa][alpha] / this->f0;
            eql2 = pqr(2) * dpqr(0) - pqr(0) * dpqr(2);
            eqr2 = pqr(2) * dpqr(1) - pqr(1) * dpqr(2);
            e += (eql1 * eql2 + eqr1 * eqr2) / pow(pqr(2), 2);
         }
      }
      this->derror[l] = 2 * e;
   }
}

// 2nd derivatives of Eq.(11.15)
void BundleAdjustment::calcddError()
{
   // Hessian w.r.t. point
   calchE();

   // Hessian w.r.t. points and camera parameters
   calchF();

   // Hessian w.r.t. camera parameters
   calchG();
}

// Hessian w.r.t. points
void BundleAdjustment::calchE()
{
   int N = this->pointN;
   double dde = 0;
   int start, end;
   int i, j;
   double c = this->c;
   for(int alpha = 0; alpha < N; alpha++){
      start = alpha * 3;
      end = start + 3;
      for(int l = start; l < end; l++){
         for(int k = l; k < end; k++){
            // compute elements of Hessian
            dde = calchEkl(l, k, alpha);

            i = l % 3;
            j = k % 3;
            //if(l == k){
            if(i == j){
               this->hE[alpha](i, j) = dde * (1.0+c);
            }else{
               this->hE[alpha](i, j) = dde;
               this->hE[alpha](j, i) = dde;
            }
         }//k
      }//l
   }//alpha


}

// 2nd derivatives of Eq.(11.15)
double BundleAdjustment::calchEkl(int k, int l, int alpha)
{
   int M = this->cameraN;
   double e = 0;
   Vector3d dpqr_k;
   Vector3d dpqr_l;
   Vector3d pqr;
   double eqr1, eql1, eqr2, eql2;

   for(int kappa = 0; kappa < M; kappa++){
      if(this->I[kappa][alpha] == 1){		// compute when there is a point `a' in the image `K'
         dpqr_k = calcdpqrak(alpha, kappa, k);
         dpqr_l = calcdpqrak(alpha, kappa, l);
         pqr << this->p[kappa][alpha], this->q[kappa][alpha], this->r[kappa][alpha];
         eql1 = pqr(2) * dpqr_k(0) - pqr(0) * dpqr_k(2);
         eqr1 = pqr(2) * dpqr_k(1) - pqr(1) * dpqr_k(2);
         eql2 = pqr(2) * dpqr_l(0) - pqr(0) * dpqr_l(2);
         eqr2 = pqr(2) * dpqr_l(1) - pqr(1) * dpqr_l(2);
         e += (eql1 * eql2 + eqr1 * eqr2) / pow(pqr(2), 4);
      }
   }
   //cout <<  "e - " <<  e << endl;
   //cin >> l;
   return e * 2.0;
}

// Hessian w.r.t. points and camera parameters.
void BundleAdjustment::calchF()
{
   int N = this->pointN;
   int N3 = N * 3;
   int M9 = this->cameraN * 9;
   int N3_M9 = N3 + M9;
   double dde = 0;
   int start, end;
   int i = 0, j = 0;
   int alpha, kappa;
   int off;

   for(alpha = 0; alpha < N; alpha++){
      this->hF[alpha] = MatrixXd::Zero(3, M9-7);
      start = alpha * 3;
      end = start + 3;
      for(int l = start; l < end; l++){// points
         i = l % 3;
         j = 0;
         for(int k = N3; k < N3_M9; k++){// Images
            off = offset(k);
            if(off == 0){
               dde = calchFkl(k, l);
               this->hF[alpha](i, j) = dde;
               j++;
            }
         }
      }
   }//alpha
}

// 2nd derivatives of Eq.(11.15)
double BundleAdjustment::calchFkl(int k, int l)
{
   int N = this->pointN;
   int N3 = N * 3;
   int M = this->cameraN;
   double e = 0;
   Vector3d dpqr_k;
   Vector3d dpqr_l;
   Vector3d pqr;
   double eqr1, eql1, eqr2, eql2;
   int alpha, kappa;
   kappa = (k - N3) / 9;
   alpha = l / 3;

   //cout << " k = " << k << " l = " << l << endl;
   if(this->I[kappa][alpha] == 1){		// compute when there is a point `a' in the image `K'
      dpqr_k = calcdpqrak(alpha, kappa, k);
      dpqr_l = calcdpqrak(alpha, kappa, l);
      pqr << this->p[kappa][alpha], this->q[kappa][alpha], this->r[kappa][alpha];
      eql1 = pqr(2)  * dpqr_k(0) - pqr(0) * dpqr_k(2);
      eqr1 = pqr(2)  * dpqr_k(1) - pqr(1) * dpqr_k(2);
      eql2 = pqr(2)  * dpqr_l(0) - pqr(0) * dpqr_l(2);
      eqr2 = pqr(2)  * dpqr_l(1) - pqr(1) * dpqr_l(2);
      e += (eql1 * eql2 + eqr1 * eqr2) / pow(pqr(2), 4);
   }
   return e * 2;
}

// Hessian w.r.t. camera parameters
void BundleAdjustment::calchG()
{
   int N = this->pointN;
   int N3 = N * 3;
   int M = this->cameraN;
   int M9 = M * 9;
   int N3_M9 = N3 + M9;
   double dde = 0;
   int start, end;
   int i = 0, j = 0;
   int l2, k2;
   int alpha, kappa;
   int off;
   double c = this->c;

   for(kappa = 0; kappa < M; kappa++){
      start = kappa * 9 + N3;
      end = start + 9;
      //i = 0;
      for(int k = start; k < end; k++){
         off = offset(k);
         if(off == 0){
            j = i;
            for(int l = k; l < end; l++){

               off = offset(l);
               if(off == 0){
                  //cout << "i = " << i << " j = " << j << endl;
                  dde = calchGkl(k, l);
                  //cout << "a" << endl;
                  if(i == j){
                     this->hG(i, j) = dde * (1.0 + c);
                  }else{
                     this->hG(i, j) = dde;
                     this->hG(j, i) = dde;
                  }
                  j++;
               }//off
            }
            i++;
         }//off
      }
   }
}

// 2nd derivatives of Eq.(11.15)
double BundleAdjustment::calchGkl(int k, int l)
{
   int N = this->pointN;
   int N3 = N * 3;
   double e = 0;
   Vector3d dpqr_k;
   Vector3d dpqr_l;
   Vector3d pqr;
   double eqr1, eql1, eqr2, eql2;
   int kappa;
   kappa = (k - N3) / 9;

   for(int alpha = 0; alpha < N; alpha++){
      if(this->I[kappa][alpha] == 1){		// compute when there is a point `a' in the image `K'
         dpqr_k = calcdpqrak(alpha, kappa, k);
         dpqr_l = calcdpqrak(alpha, kappa, l);
         pqr << this->p[kappa][alpha], this->q[kappa][alpha], this->r[kappa][alpha];
         eql1 = pqr(2) * dpqr_k(0) - pqr(0) * dpqr_k(2);
         eqr1 = pqr(2) * dpqr_k(1) - pqr(1) * dpqr_k(2);
         eql2 = pqr(2) * dpqr_l(0) - pqr(0) * dpqr_l(2);
         eqr2 = pqr(2) * dpqr_l(1) - pqr(1) * dpqr_l(2);
         e += (eql1 * eql2 + eqr1 * eqr2) / pow(pqr(2), 4);
      }
   }
   return e * 2;
}

// offset for unknown parameters
int BundleAdjustment::offset(int k)
{
   int N3 = this->pointN * 3;
   k = k - N3;
   switch (k) {
     case 3:     //t1
     case 4:     //t1
     case 5:     //t1
     case 6:     //R1
     case 7:     //R1
     case 8:     //R1
     case 13:    //t22
        return 1;
        break;
     default:
        break;
   }
   return 0;
}

// Compute tidle(E)
void BundleAdjustment::calcErrorTilde()
{
   int N = this->pointN;
   int M = this->cameraN;
   double f0 = this->f0;
   double e = 0;
   double sqrt1, sqrt2;
   Vector3d pqrTilde;

   for(int kappa = 0; kappa < M; kappa++){
      for(int alpha = 0; alpha < N; alpha++){
         if(this->I[kappa][alpha] == 1){
            pqrTilde = calcpqrTildeak(alpha, kappa);
            sqrt1 = pqrTilde(0) / pqrTilde(2) - this->x[kappa][alpha] / f0;
            sqrt2 = pqrTilde(1) / pqrTilde(2) - this->y[kappa][alpha] / f0;
            e += (sqrt1*sqrt1) + (sqrt2*sqrt2);
         }
      }
   }
   this->errorTilde = e;
}

// compute tilde(pqr)
Vector3d BundleAdjustment::calcpqrTildeak(int alpha, int kappa)
{
   double f0 = this->f0;
   Vector4d XTilde;
   double fTilde;
   double uTilde, vTilde;
   Vector3d tTilde;
   Matrix3d RTilde;

   // points
   XTilde(0) = this->X[alpha](0) + this->deltaXiP(alpha * 3 + 0);
   XTilde(1) = this->X[alpha](1) + this->deltaXiP(alpha * 3 + 1);
   XTilde(2) = this->X[alpha](2) + this->deltaXiP(alpha * 3 + 2);
   XTilde(3) = 1.0;

   // camera parameters
   fTilde = this->f[kappa] + getdeltaXiF(kappa, 0);
   uTilde = this->u[kappa] + getdeltaXiF(kappa, 1);
   vTilde = this->v[kappa] + getdeltaXiF(kappa, 2);
//  uTilde = this->u[kappa]; // + getdeltaXiF(kappa, 1);
//  vTilde = this->v[kappa]; // + getdeltaXiF(kappa, 2);

   Vector3d deltat;
   deltat << getdeltaXiF(kappa, 3), getdeltaXiF(kappa, 4), getdeltaXiF(kappa, 5);
   tTilde = this->t[kappa] + deltat;

   Vector3d deltaOmega;
   deltaOmega << getdeltaXiF(kappa, 6), getdeltaXiF(kappa, 7), getdeltaXiF(kappa, 8);
   Matrix3d deltaR;
   deltaR = getRotateMat(deltaOmega);
   RTilde = deltaR * this->R[kappa];

   // compute camera matrices
   Matrix3d KTilde;
   KTilde <<
      fTilde, 0, uTilde,
      0, fTilde, vTilde,
      0, 0, f0;
   Matrix34d ItTilde;
   Matrix34d PTilde;
   ItTilde << Matrix3d::Identity(), -tTilde;
   PTilde = KTilde * RTilde.transpose() * ItTilde;

   Vector3d pqrTilde;
   pqrTilde = PTilde * XTilde;

   return pqrTilde;
}

// for camera parameter computation
double BundleAdjustment::getdeltaXiF(int kappa, int param)
{
   double delta = 0;
   switch (kappa) {
     case 0:
        switch (param) {
          case 0:
          case 1:
          case 2:
             delta = this->deltaXiF(kappa * 9 + param);
             break;
          default:
             delta = 0;
             break;
        }
        break;
     case 1:
        switch (param) {
          case 0:
          case 1:
          case 2:
          case 3:
             delta = this->deltaXiF(kappa * 9 + param - 6);
             break;
          case 4:
             delta = 0;
             break;
          default:
             delta = this->deltaXiF(kappa * 9 - 7 + param);
             break;
        }
        break;
     default:
        delta = this->deltaXiF(kappa * 9 - 7 + param);
        break;
   }
   return delta;
}

// inline squre function
inline double sqr(double x) { return x*x; }

// for camera parameter computation
Matrix3d BundleAdjustment::getRotateMat(const Vector3d& w)
{
   Matrix3d     R;
   double       omega = w.norm();
   Vector3d     l = w;
   double       cs, sn, cs1;

   cs = cos(omega);
   sn = sin(omega);
   cs1 = 1.0 - cs;

   // for very small omega
   if (omega < EPS)
      R = I3;   // Identity matrix
   else
      R <<
         cs + sqr(l(0))*cs1, l(0)*l(1)*cs1 - l(2)*sn, l(0)*l(2)*cs1 + l(1)*sn,
         l(0)*l(1)*cs1 + l(2)*sn, cs + sqr(l(1))*cs1, l(1)*l(2)*cs1 - l(0)*sn,
         l(0)*l(2)*cs1 - l(1)*sn, l(1)*l(2)*cs1 + l(0)*sn, cs + sqr(l(2))*cs1;

   return R;
}

// Eq.(11.10)
void BundleAdjustment::renewParams()
{
   int N = this->pointN;
   int M = this->cameraN;

   // points
   int id = 0;

   for(int i = 0; i < N; i++){
      id = i * 3;
      this->X[i](0) += this->deltaXiP(id);
      this->X[i](1) += this->deltaXiP(id + 1);
      this->X[i](2) += this->deltaXiP(id + 2);
   }

   // camera parameters
   Vector3d deltat;
   Vector3d deltaOmega;
   Matrix3d deltaR;
   Matrix3d K;
   Matrix34d It;

   for(int kappa = 0; kappa < M; kappa ++){
      this->f[kappa] += getdeltaXiF(kappa, 0);

      deltat << getdeltaXiF(kappa, 3), getdeltaXiF(kappa, 4), getdeltaXiF(kappa, 5);
      this->t[kappa] = this->t[kappa] + deltat;

      deltaOmega << getdeltaXiF(kappa, 6), getdeltaXiF(kappa, 7), getdeltaXiF(kappa, 8);
      deltaR = getRotateMat(deltaOmega);
      this->R[kappa] = deltaR * this->R[kappa];
      K <<
         this->f[kappa], 0, this->u[kappa],
         0, this->f[kappa], this->v[kappa],
         0, 0, this->f0;
      It << Matrix3d::Identity(), - this->t[kappa];
      this->cameraMat[kappa] = K * this->R[kappa].transpose() * It;
   }

}


// step 2
void BundleAdjustment::solveEquations()
{
   // set delta xi_P and delta xi_F
   setdPdF();

   // solve for xi_F
   solveEquationsdXiF();

   // solve for delta X_α, delta Y_α, delta Z_α
   solveEquationsdXiP();
}

// set delta xi_P and delta xi_F
void BundleAdjustment::setdPdF()
{
   int N3 = this->pointN * 3;
   int M9 = this->cameraN * 9;
   int N3_M9 = N3 + M9;

   for(int l = 0; l < N3; l++){
      this->dP(l) = this->derror[l];
   }
   int off;
   int fl = 0;
   for(int l = N3; l < N3_M9; l++){
      off = offset(l);
      if(off == 0){
         this->dF(fl) = this->derror[l];
         fl++;
      }
   }
}

// solve for xi_F
void BundleAdjustment::solveEquationsdXiF()
{
   int N = this->pointN;
   int M9_7 = this->cameraN * 9 - 7;

   Matrix3d invEa;
   MatrixXd traFainvEa(M9_7, 3);
   MatrixXd traFainvEaFa(M9_7, M9_7);
   Vector3d nabXaE;
   VectorXd traFainvEanabXaE(M9_7);

   traFainvEaFa = MatrixXd::Zero(M9_7, M9_7);
   traFainvEanabXaE = VectorXd::Zero(M9_7);
   for(int i = 0; i < N; i++){
      invEa = this->hE[i].inverse();
      traFainvEa = this->hF[i].transpose() * invEa;

      //nabla Xa E = partial E / partial xi
      nabXaE <<
         this->derror[i * 3],
         this->derror[i * 3 + 1],
         this->derror[i * 3 + 2];
      traFainvEaFa += traFainvEa * this->hF[i];
      traFainvEanabXaE += traFainvEa * nabXaE;
   }
   MatrixXd matl(M9_7, M9_7);
   VectorXd matr(M9_7);

   // solve xi_F
   matl = this->hG - traFainvEaFa;
   matr = traFainvEanabXaE - this->dF;
   FullPivLU<MatrixXd> solver(matl);
   this->deltaXiF = solver.solve(matr);
}

// solve for delta X_α, delta Y_α, and delta Z_α
void BundleAdjustment::solveEquationsdXiP()
{
   int N = this->pointN;
   Matrix3d invEa;
   Vector3d FadXiF;
   Vector3d nabXaE;
   Vector3d dXa;

   for(int i = 0; i < N; i++){
      invEa = this->hE[i].inverse();
      FadXiF = this->hF[i] * this->deltaXiF;
      nabXaE <<
         this->derror[i * 3],
         this->derror[i * 3 + 1],
         this->derror[i * 3 + 2];

      dXa = - invEa * (FadXiF + nabXaE);
      this->deltaXiP(i * 3 + 0) = dXa(0);
      this->deltaXiP(i * 3 + 1) = dXa(1);
      this->deltaXiP(i * 3 + 2) = dXa(2);
   }
}


// allocate arrays
void BundleAdjustment::init()
{
   int N = this->pointN;
   int M = this->cameraN;
   int i, j;

   // inpute data
   this->I = new int*[M];
   this->x = new double*[M];
   this->y = new double*[M];
   for(i = 0; i < M; i++){
      this->x[i] = new double[N];
      this->y[i] = new double[N];
      this->I[i] = new int[N];
   }

   // variables
   this->X = new Vector3d[this->pointN];
   for(i = 0; i < N; i++){
      this->X[i] = Vector3d::Zero();
   }

   this->cameraMat = new Matrix34d[M];
   this->f = new double[M];
   this->u = new double[M];
   this->v = new double[M];
   this->t = new Vector3d[M];
   this->R = new Matrix3d[M];

   // errors etc.
   this->p = new double*[M];
   this->q = new double*[M];
   this->r = new double*[M];
   for(i = 0; i < M; i++){
      this->p[i] = new double[N];
      this->q[i] = new double[N];
      this->r[i] = new double[N];
   }

   int N3 = N * 3;
   int M9 = M * 9;
   int M9_7 = M9 - 7;
   int N3_M9 = N3 + M9;

   this->derror = new double[N3_M9];
   this->hE = new Matrix3d[N];
   this->hF = new MatrixXd[N];
   for(i = 0; i < N; i++){
      this->hF[i] = MatrixXd(3, M9_7);
   }
   this->hG = MatrixXd(M9_7, M9_7);
   this->deltaXiP = VectorXd(N3);
   this->deltaXiF = VectorXd(M9_7);
   this->dP = VectorXd(N3);
   this->dF = VectorXd(M9_7);
}

BundleAdjustment::BundleAdjustment()
{
   // non
}

// free
BundleAdjustment::~BundleAdjustment()
{
   int N = this->pointN;
   int M = this->cameraN;


   if(this->p != NULL){
      cout << "delete" << endl;
      for(int i = 0; i < M; i++){
         delete this->x[i];
         delete this->y[i];
         delete this->I[i];
         delete this->p[i];
         delete this->q[i];
         delete this->r[i];
      }
      delete[] this->x;
      delete[] this->y;
      delete[] this->I;
      delete[] this->p;
      delete[] this->q;
      delete[] this->r;
   }
   delete[] this->X;
   delete[] this->f;
   delete[] this->u;
   delete[] this->v;
   delete[] this->t;
   delete[] this->R;
   delete[] this->cameraMat;
   delete[] this->derror;
   delete[] this->hE;
   delete[] this->hF;


}

//
// set parameters
//
void BundleAdjustment::setX(int alpha, const Vector3d& X)
{
   this->X[alpha] = X;
}

void BundleAdjustment::setf(int kappa, double f)
{
   this->f[kappa] = f;
}

void BundleAdjustment::setu(int kappa, double u)
{
   this->u[kappa] = u;
}

void BundleAdjustment::setv(int kappa, double v)
{
   this->v[kappa] = v;
}

void BundleAdjustment::sett(int kappa, const Vector3d& t)
{
   this->t[kappa] = t;
}

void BundleAdjustment::setR(int kappa, const Matrix3d& R)
{
   this->R[kappa] = R;
}

void BundleAdjustment::setCameraMat(int kappa, const Matrix34d& P)
{
   this->cameraMat[kappa] = P;
}
