//
// Data read functions
//

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>

#include <Eigen/Dense>
#include "PerspectiveSelfCalibration.h"
#include "utils.h"

#include <stdio.h>
#include <stdlib.h>

using namespace Eigen;
using namespace std;


#define MAXBUF 8192

// read data from single file
bool
read_data(const char filename[], MatrixXd **pt0, int *CamNum, int *PtNum)
{
   MatrixXd     *ptr;
   std::ifstream ifs(filename);

   if (ifs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   char buff[MAXBUF];

   // read # of cameras
   while (ifs.getline(buff, MAXBUF-1))
   {
      if (buff[0] == '#')
         continue;
      else
         break;
   }
   if (!sscanf(buff, "%d", CamNum) == 1)
      return false;
   
   // read # of points
   while (ifs.getline(buff, MAXBUF-1))
   {
      if (buff[0] == '#')
         continue;
      else
         break;
   }
   if (!sscanf(buff, "%d", PtNum) == 1)
      return false;

   // data allocation
   ptr = *pt0 = new MatrixXd[*PtNum];

   // skip comments
   while (ifs.getline(buff, MAXBUF-1))
   {
      if (buff[0] == '#')
         continue;
      else
         break;
   }

   // read data
   for (int p = 0; p < *PtNum; p++)
   {
      char *strp, *nstrp;
      
      ptr[p].resize(2,*CamNum);

      // copy buff
      strp = buff;
      
      for (int c = 0; c < *CamNum; c++)
      {
         double x0, y0;
         
         // x
         x0 = strtod(strp, &nstrp);
         if (*nstrp != '\0')
         {
            ptr[p].col(c)(0) = x0;
            strp = nstrp;
         }
         else
            return false;

         // y
         y0 = strtod(strp, &nstrp);
         if (*nstrp != '\0')
         {
            ptr[p].col(c)(1) = y0;
            strp = nstrp;
         }
         else
            return false;
      }
      // read next line
      ifs.getline(buff, MAXBUF-1);
   }

   return true;
}

// read data from multiple files
bool
read_data_files(const char filetemplate[], int CamNum, MatrixXd **pt0, int *PtNum)
{
   MatrixXd     *ptr;
   char         filename[MAXBUF];
   char         buff[MAXBUF];
   int          num;
   std::ifstream ifs;
   double       x0, y0;

   for (int cam = 0; cam < CamNum; cam++)
   {
      sprintf(filename, filetemplate, cam);
      ifs.open(filename);

//      std::cerr << "Open file: " << filename << std::endl;

      if (ifs.fail())
      {
         std::cerr << "Cannot open file:" << filename << std::endl;
         return false;
      }

      // read # of points
      ifs.getline(buff, MAXBUF-1);
      if (sscanf(buff, "%d", &num) != 1)
      {
         std::cerr << "There is no number of points in "
                   << filename << std::endl;
         return false;
      }

      // data allocation
      if (cam == 0)
      {
         *PtNum = num;
         ptr = *pt0 = new MatrixXd[*PtNum];
         for (int p = 0; p < *PtNum; p++)
            ptr[p].resize(2, CamNum);
      }
      else if (num != *PtNum)
      {
         std::cerr << "Data number mismatched in "
                   << filename << std::endl;
         return false;
      }

      // read data from each file
      for (int p = 0; p < *PtNum; p++)
      {
         ifs.getline(buff, MAXBUF-1);

         // skip blank line
         while (!strlen(buff))
            ifs.getline(buff, MAXBUF-1);
            
         if (sscanf(buff, "%lf %lf", &x0, &y0) != 2)
         {
            std::cerr << "Point data format is wrong in "
                      << filename << std::endl;
            return false;
         }

         ptr[p].col(cam)(0) = x0;
         ptr[p].col(cam)(1) = y0;
      }

      ifs.close();
   }

   return true;
}

// write data to file
bool
write_data(const char filename[], const Vector4d X[], int PtNum)
{
   std::ofstream ofs(filename);

   if (ofs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   // output the number of data
   ofs << PtNum << std::endl;

   // set output format
   ofs << scientific << setprecision(15);

   // output data
   for (int i = 0; i < PtNum; i++)
      ofs << X[i](0)/X[i](3) << " "
          << X[i](1)/X[i](3) << " "
          << X[i](2)/X[i](3) << std::endl;

   return true;
}

// write data to file
bool
write_3d_data(const char filename[], const Vector3d X[], int PtNum)
{
   std::ofstream ofs(filename);

   if (ofs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   // output the number of data
   ofs << PtNum << std::endl;

   // set output format
   ofs << scientific << setprecision(15);

   // output data
   for (int i = 0; i < PtNum; i++)
      ofs << X[i](0) << " "
          << X[i](1) << " "
          << X[i](2) << std::endl;

   return true;
}

// write data to file
bool
write_4d_data(const char filename[], const Vector4d X[], int PtNum)
{
   std::ofstream ofs(filename);

   if (ofs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   // output the number of data
   ofs << PtNum << std::endl;

   // set output format
   ofs << scientific << setprecision(15);

   // output data
   for (int i = 0; i < PtNum; i++)
      ofs << X[i](0) << " "
          << X[i](1) << " "
          << X[i](2) << " "
          << X[i](3) << std::endl;

   return true;
}

// write projection matrices to files
bool
write_proj_files(const char filetemplate[], const Matrix34d Proj[], int CamNum)
{
   char         filename[MAXBUF];
   std::ofstream ofs;
   
   for (int cam = 0; cam < CamNum; cam++)
   {
      sprintf(filename, filetemplate, cam);
      ofs.open(filename);

      if (ofs.fail())
      {
         std::cerr << "Cannot open file:" << filename << std::endl;
         return false;
      }

      ofs << scientific << setprecision(15);
      ofs << Proj[cam];

      ofs.close();
   }

   return true;
}

// write ply format output
bool
write_ply_data(const char filename[], const Vector4d X[], int PtNum,
               int r, int g, int b)
{
   std::ofstream ofs(filename);

   if (ofs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   // wirte headers
   ofs << "ply" << endl
       << "format ascii 1.0" << endl
       << "comment created by 3d_to_ply" << endl
       << "element vertex " << PtNum << endl
       << "property float32 x" << endl
       << "property float32 y" << endl
       << "property float32 z" << endl
       << "property uint8 red" << endl
       << "property uint8 green" << endl
       << "property uint8 blue" << endl
       << "end_header" << endl;

   // output data
   for (int i = 0; i < PtNum; i++)
   {
      char buff[256];

      sprintf(buff, "%lf %lf %lf %d %d %d",
	      X[i](0)/X[i](3), X[i](1)/X[i](3), X[i](2)/X[i](3), r, g, b);
      ofs << buff << endl;
   }
   
   return true;
}

// write ply format output
bool
write_3d_ply_data(const char filename[], const Vector3d X[], int PtNum,
                  int r, int g, int b)
{
   std::ofstream ofs(filename);

   if (ofs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   // wirte headers
   ofs << "ply" << endl
       << "format ascii 1.0" << endl
       << "comment created by 3d_to_ply" << endl
       << "element vertex " << PtNum << endl
       << "property float32 x" << endl
       << "property float32 y" << endl
       << "property float32 z" << endl
       << "property uint8 red" << endl
       << "property uint8 green" << endl
       << "property uint8 blue" << endl
       << "end_header" << endl;

   // output data
   for (int i = 0; i < PtNum; i++)
   {
      char buff[256];

      sprintf(buff, "%lf %lf %lf %d %d %d",
	      X[i](0), X[i](1), X[i](2), r, g, b);
      ofs << buff << endl;
   }
   
   return true;
}
