//
// Data read functions
//

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>

#include "PlaneReconstruction.h"
#include "utils.h"

using namespace Eigen;
using namespace std;

#define MAXBUF 256
#define MAXLEN 1024

bool
read_Hmat(char filename[], Matrix3d& F)
{
   std::ifstream ifs(filename);

   if (ifs.fail())
   {
      std::cerr << "Cannot open file: " << filename << std::endl;
      return false;
   }

   char buff[MAXBUF];
   double f[9];
   int n;

   ifs.getline(buff, MAXBUF-1);
   n = sscanf(buff, "%lf %lf %lf", f, f+1, f+2);
   if (n != 3) return false;

   ifs.getline(buff, MAXBUF-1);
   n = sscanf(buff, "%lf %lf %lf", f+3, f+4, f+5);
   if (n != 3) return false;

   ifs.getline(buff, MAXBUF-1);
   n = sscanf(buff, "%lf %lf %lf", f+6, f+7, f+8);
   if (n != 3) return false;

   F << f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7], f[8];

   return true;
}

bool
read_Hmat(char filename[], Vector9d& theta)
{
   std::ifstream ifs(filename);

   if (ifs.fail())
   {
      std::cerr << "Cannot open file: " << filename << std::endl;
      return false;
   }

   char buff[MAXBUF];
   double f[9];
   int n;

   ifs.getline(buff, MAXBUF-1);
   n = sscanf(buff, "%lf %lf %lf %lf %lf %lf %lf %lf %lf",
              f, f+1, f+2, f+3, f+4, f+5, f+6, f+7, f+8);
   if (n != 9) return false;

   theta << f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7], f[8];

   return true;
}

bool
read_data(char filename[], Vector2d **pos0, int *Num)
{
   std::ifstream ifs(filename);

   if (ifs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   char buff[MAXBUF];
   int n;

   ifs.getline(buff, MAXBUF-1);
   n = sscanf(buff, "%d", Num);
   if (n != 1) return false;

   *pos0 = new Vector2d [*Num];

   for (int i = 0; i < *Num; i++)
   {
      double x0, y0;
      if (ifs.eof())  return false;
      ifs.getline(buff, MAXBUF-1);
      n = sscanf(buff, "%lf %lf", &x0, &y0);
      if (n == EOF)
      { // null line
         i--;
         continue;
      }
      else if (n != 2) return false;
      (*pos0)[i] << x0, y0;
   }

   return true;
}

bool
read_data(char filename[], double **x0, double **y0, int *Num)
{
   std::ifstream ifs(filename);

   if (ifs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   char buff[MAXBUF];
   int n;

   ifs.getline(buff, MAXBUF-1);
   n = sscanf(buff, "%d", Num);
   if (n != 1) return false;

   *x0 = new double [*Num];
   *y0 = new double [*Num];

   for (int i = 0; i < *Num; i++)
   {
      if (ifs.eof())  return  false;
      ifs.getline(buff, MAXBUF-1);
      n = sscanf(buff, "%lf %lf", (*x0 + i), (*y0 + i));
      if (n == EOF)
      { // null line
         i--;
         continue;
      }
      else if (n != 2) return false;
   }

   return true;
}


bool
read_data(char filename[], double **x0, double **y0, double **x1, double **y1,
          int *Num)
{
   std::ifstream ifs(filename);

   if (ifs.fail())
   {
      std::cerr << "Cannot open file" << std::endl;
      return false;
   }

   char buff[MAXBUF];
   int n;

   ifs.getline(buff, MAXBUF-1);
   n = sscanf(buff, "%d", Num);
   if (n != 1) return false;

   *x0 = new double [*Num];
   *y0 = new double [*Num];
   *x1 = new double [*Num];
   *y1 = new double [*Num];

   for (int i = 0; i < *Num; i++)
   {
      if (ifs.eof())  return  false;
      ifs.getline(buff, MAXBUF-1);
      n = sscanf(buff, "%lf %lf %lf %lf",
                 (*x0 + i), (*y0 + i), (*x1 + i), (*y1 + i));
      if (n == EOF)
      { // null line
         i--;
         continue;
      }
      if (n != 4) return false;
   }

   return true;
}

bool
write_data(char filename[], Vector2d *pos0, int Num)
{
   std::ofstream ofs(filename);

   if (ofs.fail())
   {
      std::cerr << "Cannot open file:" << filename << std::endl;
      return false;
   }

   // output the number of data
   ofs << Num << std::endl;

   // set output format
   ofs << scientific << setprecision(15);

   // output data
   for (int i = 0; i < Num; i++)
      ofs << pos0[i](0) << " " << pos0[i](1) << std::endl;

   return true;
}

// output 3-D
bool
write_3D_data(char fname[], Vector3d X[], int Num)
{
   ofstream ofs(fname);

   if (ofs.fail())
   {
      cerr << "File '" << fname << "' cannot create!\n";
      return false;
   }

   ofs << Num << endl;

   ofs << setprecision(15);
   ofs << setiosflags(ios::scientific);

   for (int i = 0; i < Num; i++)
   {
      ofs << X[i](0) << " "
	  << X[i](1) << " "
	  << X[i](2) << endl;
   }

   return true;
}

// 3Dデータをply形式で出力
bool
write_ply_data(char plyname[],
               Vector3d X[], int Num,
               int r, int g, int b)
{
   ofstream ofs(plyname);

   if (ofs.fail())
   {
      cerr << "File '" << plyname << "' cannot create!\n";
      return false;
   }

   // wirte headers
   ofs << "ply" << endl
       << "format ascii 1.0" << endl
       << "comment created by simtest3" << endl
       << "element vertex " << Num << endl
       << "property float32 x" << endl
       << "property float32 y" << endl
       << "property float32 z" << endl
       << "property uint8 red" << endl
       << "property uint8 green" << endl
       << "property uint8 blue" << endl
       << "end_header" << endl;

   for (int i = 0; i < Num; i++)
   {
      char buff[MAXLEN];

      sprintf(buff, "%lf %lf %lf %d %d %d",
	      X[i](0), X[i](1), X[i](2), r, g, b);
      ofs << buff << endl;
   }

   return true;
}
